@@ -10,21 +10,21 @@
 %                             UUU   R  R   LLLLL                              %
 %                                                                             %
 %                                                                             %
-%                        Retrieve An Image Via a URL.                         %
+%                        Retrieve An Image Via URL.                           %
 %                                                                             %
 %                              Software Design                                %
-%                                John Cristy                                  %
+%                                   Cristy                                    %
 %                              Bill Radcliffe                                 %
 %                                March 2000                                   %
 %                                                                             %
 %                                                                             %
-%  Copyright 1999-2008 ImageMagick Studio LLC, a non-profit organization      %
+%  Copyright @ 1999 ImageMagick Studio LLC, a non-profit organization         %
 %  dedicated to making software imaging solutions freely available.           %
 %                                                                             %
 %  You may not use this file except in compliance with the License.  You may  %
 %  obtain a copy of the License at                                            %
 %                                                                             %
-%    http://www.imagemagick.org/script/license.php                            %
+%    https://imagemagick.org/script/license.php                               %
 %                                                                             %
 %  Unless required by applicable law or agreed to in writing, software        %
 %  distributed under the License is distributed on an "AS IS" BASIS,          %
@@ -40,33 +40,29 @@
 /*
   Include declarations.
 */
-#include "magick/studio.h"
-#include "magick/blob.h"
-#include "magick/blob-private.h"
-#include "magick/constitute.h"
-#include "magick/exception.h"
-#include "magick/exception-private.h"
-#include "magick/image.h"
-#include "magick/image-private.h"
-#include "magick/list.h"
-#include "magick/magick.h"
-#include "magick/memory_.h"
-#include "magick/quantum-private.h"
-#include "magick/static.h"
-#include "magick/resource_.h"
-#include "magick/string_.h"
-#include "magick/module.h"
-#if defined(MAGICKCORE_XML_DELEGATE)
-#  if defined(__WINDOWS__)
-#    if defined(__MINGW32__)
-#      define _MSC_VER
-#    endif
-#    include <win32config.h>
+#include "MagickCore/studio.h"
+#include "MagickCore/blob.h"
+#include "MagickCore/blob-private.h"
+#include "MagickCore/constitute.h"
+#include "MagickCore/delegate.h"
+#include "MagickCore/exception.h"
+#include "MagickCore/exception-private.h"
+#include "MagickCore/image.h"
+#include "MagickCore/image-private.h"
+#include "MagickCore/list.h"
+#include "MagickCore/magick.h"
+#include "MagickCore/memory_.h"
+#include "MagickCore/module.h"
+#include "MagickCore/quantum-private.h"
+#include "MagickCore/static.h"
+#include "MagickCore/resource_.h"
+#include "MagickCore/string_.h"
+#include "MagickCore/utility.h"
+#if defined(MAGICKCORE_WINDOWS_SUPPORT)
+#  include <urlmon.h>
+#  if !defined(__MINGW32__)
+#    pragma comment(lib, "urlmon.lib")
 #  endif
-#  include <libxml/parser.h>
-#  include <libxml/xmlmemory.h>
-#  include <libxml/nanoftp.h>
-#  include <libxml/nanohttp.h>
 #endif
 
 /*
@@ -80,7 +76,7 @@
 %                                                                             %
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %
-%  ReadURLImage retrieves an image via a URL, decodes the image, and returns
+%  ReadURLImage retrieves an image via URL, decodes the image, and returns
 %  it.  It allocates the memory necessary for the new Image structure and
 %  returns a pointer to the new image.
 %
@@ -90,46 +86,54 @@
 %
 %  A description of each parameter follows:
 %
-%    o image_info: The image info.
+%    o image_info: the image info.
 %
 %    o exception: return any errors or warnings in this structure.
 %
-%
 */
 
-#if defined(__cplusplus) || defined(c_plusplus)
-extern "C" {
-#endif
-
-static void GetFTPData(void *userdata,const char *data,int length)
+#if !defined(MAGICKCORE_WINDOWS_SUPPORT)
+static Image* InvokeURLDelegate(ImageInfo* read_info,Image* image,
+  const char *delegate,ExceptionInfo* exception)
 {
-  FILE
-    *file;
+  Image
+    *images,
+    *next;
 
-  file=(FILE *) userdata;
-  if (file == (FILE *) NULL)
-    return;
-  if (length <= 0)
-    return;
-  (void) fwrite(data,length,1,file);
-}
+  MagickBooleanType
+    status;
 
-#if defined(__cplusplus) || defined(c_plusplus)
+  images=(Image *) NULL;
+  status=InvokeDelegate(read_info,image,delegate,(char *) NULL,
+    exception);
+  if (status != MagickFalse)
+    {
+      (void) FormatLocaleString(read_info->filename,MagickPathExtent,
+        "%s.dat",read_info->unique);
+      *read_info->magick='\0';
+      images=ReadImage(read_info,exception);
+      (void) RelinquishUniqueFileResource(read_info->filename);
+      if (images != (Image *) NULL)
+        for (next=images; next != (Image *) NULL; next=next->next)
+          (void) CopyMagickString(next->filename,image->filename,
+            MagickPathExtent);
+    }
+  return(images);
 }
 #endif
 
 static Image *ReadURLImage(const ImageInfo *image_info,ExceptionInfo *exception)
 {
-#define MaxBufferExtent  8192
-
   char
-    filename[MaxTextExtent];
+    filename[MagickPathExtent];
 
   FILE
     *file;
 
   Image
-    *image;
+    *image,
+    *images,
+    *next;
 
   ImageInfo
     *read_info;
@@ -137,87 +141,84 @@ static Image *ReadURLImage(const ImageInfo *image_info,ExceptionInfo *exception)
   int
     unique_file;
 
-  image=(Image *) NULL;
+  images=(Image *) NULL;
+  image=AcquireImage(image_info,exception);
   read_info=CloneImageInfo(image_info);
   SetImageInfoBlob(read_info,(void *) NULL,0);
+#if !defined(MAGICKCORE_WINDOWS_SUPPORT)
+  if (LocaleCompare(read_info->magick,"http") == 0)
+    {
+      images=InvokeURLDelegate(read_info,image,"http:decode",exception);
+      read_info=DestroyImageInfo(read_info);
+      image=DestroyImage(image);
+      return(images);
+    }
+  if (LocaleCompare(read_info->magick,"https") == 0)
+    {
+      images=InvokeURLDelegate(read_info,image,"https:decode",exception);
+      read_info=DestroyImageInfo(read_info);
+      image=DestroyImage(image);
+      return(images);
+    }
+#endif
+  if (LocaleCompare(read_info->magick,"file") == 0)
+    {
+      (void) CopyMagickString(read_info->filename,image_info->filename+2,
+        MagickPathExtent);
+      *read_info->magick='\0';
+      images=ReadImage(read_info,exception);
+      read_info=DestroyImageInfo(read_info);
+      image=DestroyImage(image);
+      return(GetFirstImageInList(images));
+    }
   file=(FILE *) NULL;
   unique_file=AcquireUniqueFileResource(read_info->filename);
   if (unique_file != -1)
     file=fdopen(unique_file,"wb");
   if ((unique_file == -1) || (file == (FILE *) NULL))
     {
-      read_info=DestroyImageInfo(read_info);
-      (void) CopyMagickString(image->filename,read_info->filename,
-        MaxTextExtent);
       ThrowFileException(exception,FileOpenError,"UnableToCreateTemporaryFile",
-        image->filename);
-      image=DestroyImageList(image);
+        read_info->filename);
+      read_info=DestroyImageInfo(read_info);
+      image=DestroyImage(image);
       return((Image *) NULL);
     }
-  (void) CopyMagickString(filename,image_info->magick,MaxTextExtent);
-  (void) ConcatenateMagickString(filename,":",MaxTextExtent);
+  (void) CopyMagickString(filename,image_info->magick,MagickPathExtent);
+  (void) ConcatenateMagickString(filename,":",MagickPathExtent);
   LocaleLower(filename);
-  (void) ConcatenateMagickString(filename,image_info->filename,MaxTextExtent);
-  if (LocaleCompare(read_info->magick,"file") == 0)
-    {
-      (void) RelinquishUniqueFileResource(read_info->filename);
-      unique_file=(-1);
-      (void) CopyMagickString(read_info->filename,image_info->filename+2,
-        MaxTextExtent);
-    }
-#if defined(MAGICKCORE_XML_DELEGATE) && defined(LIBXML_FTP_ENABLED)
-  if (LocaleCompare(read_info->magick,"ftp") == 0)
-    {
-      void
-        *context;
-
-      xmlNanoFTPInit();
-      context=xmlNanoFTPNewCtxt(filename);
-      if (context != (void *) NULL)
-        {
-          if (xmlNanoFTPConnect(context) >= 0)
-            (void) xmlNanoFTPGet(context,GetFTPData,(void *) file,
-              (char *) NULL);
-          (void) xmlNanoFTPClose(context);
-        }
-    }
-#endif
-#if defined(MAGICKCORE_XML_DELEGATE) && defined(LIBXML_HTTP_ENABLED)
-  if (LocaleCompare(read_info->magick,"http") == 0)
-    {
-      char
-        buffer[MaxBufferExtent],
-        *type;
-
-      int
-        bytes;
-
-      void
-        *context;
-
-      type=(char *) NULL;
-      context=xmlNanoHTTPMethod(filename,(const char *) NULL,
-        (const char *) NULL,&type,(const char *) NULL,0);
-      if (context != (void *) NULL)
-        {
-          while ((bytes=xmlNanoHTTPRead(context,buffer,MaxBufferExtent)) > 0)
-            (void) fwrite(buffer,bytes,1,file);
-          xmlNanoHTTPClose(context);
-          xmlFree(type);
-          xmlNanoHTTPCleanup();
-        }
-    }
-#endif
+  (void) ConcatenateMagickString(filename,image_info->filename,
+    MagickPathExtent);
+#if defined(MAGICKCORE_WINDOWS_SUPPORT)
   (void) fclose(file);
+  if (URLDownloadToFile(NULL,filename,read_info->filename,0,NULL) != S_OK)
+    {
+      ThrowFileException(exception,FileOpenError,"UnableToOpenFile",
+        filename);
+      (void) RelinquishUniqueFileResource(read_info->filename);
+      read_info=DestroyImageInfo(read_info);
+      image=DestroyImage(image);
+      return((Image *) NULL);
+    }
+#else
+  (void) fclose(file);
+#endif
   *read_info->magick='\0';
-  image=ReadImage(read_info,exception);
-  if (unique_file != -1)
-    (void) RelinquishUniqueFileResource(read_info->filename);
+  images=ReadImage(read_info,exception);
+  (void) RelinquishUniqueFileResource(read_info->filename);
+  if (images != (Image *) NULL)
+    for (next=images; next != (Image *) NULL; next=next->next)
+      (void) CopyMagickString(next->filename,image->filename,MagickPathExtent);
   read_info=DestroyImageInfo(read_info);
-  if (image == (Image *) NULL)
-    (void) ThrowMagickException(exception,GetMagickModule(),CoderError,
-      "NoDataReturned","`%s'",filename);
-  return(GetFirstImageInList(image));
+  image=DestroyImage(image);
+  if (images != (Image *) NULL)
+    GetPathComponent(image_info->filename,TailPath,images->filename);
+  else
+    {
+      (void) ThrowMagickException(exception,GetMagickModule(),CoderError,
+        "NoDataReturned","`%s'",filename);
+      return((Image *) NULL);
+    }
+  return(GetFirstImageInList(images));
 }
 
 /*
@@ -240,35 +241,31 @@ static Image *ReadURLImage(const ImageInfo *image_info,ExceptionInfo *exception)
 %
 %  The format of the RegisterURLImage method is:
 %
-%      unsigned long RegisterURLImage(void)
+%      size_t RegisterURLImage(void)
 %
 */
-ModuleExport unsigned long RegisterURLImage(void)
+ModuleExport size_t RegisterURLImage(void)
 {
   MagickInfo
     *entry;
 
-  entry=SetMagickInfo("HTTP");
-#if defined(MAGICKCORE_XML_DELEGATE) && defined(LIBXML_HTTP_ENABLED)
+  entry=AcquireMagickInfo("URL","HTTP","Uniform Resource Locator (http://)");
+  entry->decoder=(DecodeImageHandler *) ReadURLImage;
+  entry->format_type=ImplicitFormatType;
+  (void) RegisterMagickInfo(entry);
+  entry=AcquireMagickInfo("URL","HTTPS","Uniform Resource Locator (https://)");
+  entry->decoder=(DecodeImageHandler *) ReadURLImage;
+  entry->format_type=ImplicitFormatType;
+  (void) RegisterMagickInfo(entry);
+  entry=AcquireMagickInfo("URL","FTP","Uniform Resource Locator (ftp://)");
+#if defined(MAGICKCORE_WINDOWS_SUPPORT)
   entry->decoder=(DecodeImageHandler *) ReadURLImage;
 #endif
-  entry->description=ConstantString("Uniform Resource Locator (http://)");
-  entry->module=ConstantString("URL");
-  entry->stealth=MagickTrue;
+  entry->format_type=ImplicitFormatType;
   (void) RegisterMagickInfo(entry);
-  entry=SetMagickInfo("FTP");
-#if defined(MAGICKCORE_XML_DELEGATE) && defined(LIBXML_FTP_ENABLED)
+  entry=AcquireMagickInfo("URL","FILE","Uniform Resource Locator (file://)");
   entry->decoder=(DecodeImageHandler *) ReadURLImage;
-#endif
-  entry->description=ConstantString("Uniform Resource Locator (ftp://)");
-  entry->module=ConstantString("URL");
-  entry->stealth=MagickTrue;
-  (void) RegisterMagickInfo(entry);
-  entry=SetMagickInfo("FILE");
-  entry->decoder=(DecodeImageHandler *) ReadURLImage;
-  entry->description=ConstantString("Uniform Resource Locator (file://)");
-  entry->module=ConstantString("URL");
-  entry->stealth=MagickTrue;
+  entry->format_type=ImplicitFormatType;
   (void) RegisterMagickInfo(entry);
   return(MagickImageCoderSignature);
 }
